!------------------------------------------------------------------
! subroutine to find the connectivities and generate the lwork
! files for parallel computing
!
! Variables:
!   GNNum : stores global nodal numbering of all local nodes
!           '-1' was assigned for empty entities..
!   RelNd : relative nodal information
!           e.g. RelNd(4,3,2)=1 means the global nodal numbering
!                of proc4 and proc3's second overlapping node is 1
!   CoMap: connectivity map
!           e.g CoMap(4,3) = 1 and CoMap(3,4) = 0 means
!               proc4 is master(receive) and proc3 is slave(send)
!               CoMap(:,:) = -1 means no link between two proc's
!   NOvNd : Numbers of Overlapping Nodes
!   RelPr : Relative processor No.
!
! NOTE: NP-1 is the maximum possible tasks
!------------------------------------------------------------------
subroutine gen_lwork(eps,NP, LNNod, xg, IPER, ConNP, &
                     MCP, NCP, OCP)

  integer,      intent(in) :: NP, LNNod(:), IPER(:,:), ConNP(:,:), &
                            MCP(:), NCP(:), OCP(:)
  real(kind=8), intent(in) :: xg(:,:,:)

  integer :: i, j, k, l, m, n, ii, jj, mm, nn
  integer :: i1, i2, i3, j1, j2, j3  
  integer :: flag, ier, UMF(NP), Pflag(NP)
  integer :: RelNd(NP,NP,MAXVAL(LNNod))
  integer :: NOvNd(NP,NP), CoMap(NP,NP)
  integer :: Ntask, p, datalen, msgnum, msglen, RelPr
  integer :: temp1((NP-1),NP), temp2(NP,NP)
  integer :: msgtag((NP-1),NP), itmp

  real(kind=8) :: eps, xm(3),xn(3), xgold(NP,MAXVAL(LNNod),3), dist

  integer, allocatable :: task(:,:)
  character(len=30)    :: fname
  character(len=10)    :: cname
  
  eps = 1.0d-3*eps

  RelNd = -1
  CoMap = -1
  NOvNd = 0     ! Common nodes between proc m,n
  msgtag = 100

  xgold = 10000.0d0



  ! loop through all the processors, starting from NP
  do m = NP, 2, -1   ! master m
    write(*,*) '!== working on master processor:', m, ' ==!'
    l = 0  ! Total master boundary nodes 
    
    
    ! Loop trhough boundary nodes only
    j = 0
    itmp = 0
    do j3 = 1, OCP(m)
    do j2 = 1, NCP(m)
    do j1 = 1, MCP(m)
    j = j + 1
    
    if (j3==1 .or. j3==OCP(m) .or. &
        j2==1 .or. j2==NCP(m) .or. &
        j1==1 .or. j1==MCP(m)) then
 
      xm = xg(j,:,m)
      itmp = itmp +1
      
      write(*,*) j, "/",LNNod(m)

      ! Loop over slaves n, proc smaller then master m
      do n = m-1, 1, -1   
      
        ! check if master and slave overlap
        if (ConNP(m,n) == 1) then
	write(*,*) '!== working on master processor:', n, ' ==!'
	
        ! Loop trhough boundary nodes only
        i = 0
        do i3 = 1, OCP(n)
        do i2 = 1, NCP(n)
        do i1 = 1, MCP(n)
        i = i + 1
        if (i3==1 .or. i3==OCP(n) .or. &
            i2==1 .or. i2==NCP(n) .or. &
            i1==1 .or. i1==MCP(n)) then

          xn = xg(i,:,n) 
            
          ! check 'proc m-1' with 'proc m' for overlapping nodes
          ! check if their distance < eps, IPER(i) == i (master)
          dist = sqrt(sum((xn-xm)*(xn-xm)))
	    
          if (dist < eps .and. IPER(i,n) == i .and. IPER(j,m) == j) then

            
	      
            ! check if the overlapping node belongs to
            ! higher order processors
            flag = 0
            loop1: do ii = m, NP  ! m+1 means higher order master
                ! check only if two processors meet
			 !write(*,*) m,n,ii
			
			
              if ((ConNP(ii,n) == 1)) then
                do jj = 1, LNNod(ii)

         dist = sqrt(sum((xn-xgold(ii,jj,:))*(xn-xgold(ii,jj,:))))            
                  if (dist < eps) then

                    flag = 1
                    exit loop1
                  end if
                enddo
              endif
            enddo  loop1           

            ! store the overlapping node
            if (flag == 0) then
              NOvNd(m,n) = NOvNd(m,n) + 1
	      NOvNd(n,m) = NOvNd(m,n)
	      
              l = l + 1	      
	      
              RelNd(m,n,NOvNd(m,n)) = j
              RelNd(n,m,NOvNd(m,n)) = i
            

              xgold(m,l,:) = xn ! store the overlap location

              ! 'proc m' is master and 'proc n' is slave
              CoMap(m,n) = 1
              CoMap(n,m) = 0
            endif
	    
          endif
	  

        endif ! Slave nodes
        enddo
        enddo
        enddo 
	
        endif	! Proc overlap   
      enddo     ! Slave procs 
    
    
    endif       ! Master nodes 
    enddo
    enddo
    enddo
  enddo         ! Master procs
 
  ! find the task info and total number of tasks
  ! always check through the largest to smallest proc ID
  temp2 = CoMap    ! for tmp CoMap
  temp1 = 0        ! for tmp task
  k     = 0
  do m = NP, 2, -1       ! master
    do n = m-1, 1, -1    ! slave
      Pflag = 1
      if (temp2(m,n) == 1) then
        Pflag(m) = 0
        Pflag(n) = 0
        k = k + 1

        temp1(k,m) = n   ! setup the task link
        temp1(k,n) = m
        temp2(m,n) = -1  ! if the link has been visited
        temp2(n,m) = -1  ! assign '-1' to the map


        ! check through the lower order processor ID
        ! since m is visited, starting from m-1
        do mm = m-1, 2, -1
          do nn = mm-1, 1, -1
            if (Pflag(mm) /= 0 .and. Pflag(nn) /= 0) then
              if (temp2(mm,nn) == 1) then
                Pflag(mm) = 0      ! if the proc has already been
                Pflag(nn) = 0      ! vivisted, assign '0'
                temp1( k,mm) = nn  ! for avaliable links
                temp1( k,nn) = mm  ! assign same task number
                temp2(mm,nn) = -1
                temp2(nn,mm) = -1
  
              end if
            end if
          end do
        end do
        
      end if
      
    end do
  end do

  Ntask = k

  ! write out the task for checking..
  !do i = 1, Ntask
  !  write(*,*) (temp1(i,j),j = NP, 1, -1)
  !end do

  allocate(task(Ntask,NP), stat=ier)
  if(ier /= 0) stop 'Allocation Error in "gen_lwork"'

  ! initialize task with '-1
  task = -1

  ! assign task from temp1 with correct dimensions
  task(:,:) = temp1(1:Ntask,:)

  !do i = 1, Ntask
  !  write(*,*) (task(i,j),j=1,NP)
  !end do

  ! loop through all processors to write the lwork files
  do p = 1, NP

    UMF(p) = 30 + p
    fname = trim('lwork'// trim(cname(p))//'.dat')
    open(UMF(p), file = fname, status = 'unknown')

    ! calculate the data length
    datalen = 1
    do i = 1, Ntask
      datalen = datalen + 4
      if (task(i,p) /= 0) then
        do j = 1, NOvNd(p,task(i,p))
          datalen = datalen + 2
        end do
      end if
    end do

    msgnum  = 100  ! message number
    msglen  = 1    ! message length
    write(UMF(p),210) Ntask, datalen
    do i = 1, Ntask
      RelPr = task(i,p)
      if (RelPr == 0) then
        write(UMF(p),220) msgnum
        write(UMF(p),220) 0
        write(UMF(p),220) -1
        write(UMF(p),220) 0
      else
        write(UMF(p),220) msgtag(i,p)
!        write(UMF(p),220) msgnum            ! message number
        write(UMF(p),220) CoMap(p,RelPr)    ! send or recv
        write(UMF(p),220) RelPr             ! processor ID
        write(UMF(p),220) NOvNd(p,RelPr)    ! number of msg's
        do j = 1, NOvNd(p,RelPr)
!          write(UMF(p),220) G2LNo(RelNd(p,RelPr,j),p)
          write(UMF(p),220) RelNd(p,RelPr,j)
          write(UMF(p),220) msglen
        end do
      end if
    end do
  end do
  
  210 format (2I8)
  220 format (I8)

  do i = 1, NP
    close(UMF(i))
  end do

  deallocate(task)

end subroutine gen_lwork
